/* plotsubs.c -- low level plotting routines (plots, plot, factor, etc.) to
replace FORTRAN versions. This should improve I/O performance and flexibility.
Doug A. McNaught, Feb. 1989 */
/* Contains the following subroutines:
   PLOTS   Initialize plot
   PLOT    Move the pen
   WHERE   Find pen location
   POLY    Draw a (possibly filled) polygon
   SHADE   Draw a curge with changing gray levels
   RASTER
   CRASTER
   NEWPEN  Change graphics pen
   ADVPAG  Go to a new page
   ADVPLT  Go to a new plot frame
   FACTOR  Set plot scaling factor
   GINPUT  Get location of crosshairs (interactive only)
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

/* Remove the following definition if inappropriate */
#define ENVIRON "PLOTHEADER"

/* First, define static variables equivalent to the /pltdta/ common block.
   They will be visible only to the routines in this file. */

static double fact,xx,yy; /* Scaling factor and current location */
static FILE *plotfile; /* File to receive output */

static int blank(text,len)
char text[]; 
int len;
/* Tests whether a FORTRAN string is made of only blanks. Does not require
   a terminating null. */
{
   int i,val;
   
   val = 1; /* TRUE */
   if(len != 0)
     for(i=0;i<len;i++)
       if(!isspace(text[i])) val = 0; /* FALSE */
   return(val);
}

plots_(idum,jdum,idev,name,nlen)
int *idum,*jdum,*idev; /* Addresses of integer parameters */
char *name; /* Address of FORTRAN character variable */
int nlen; /* Length of FORTRAN string (NOT passed by reference) */

/* Initializes plotting into a PostScript metafile. The first three arguments 
   are present for historical reasons only and are ignored, except for idum,
   which is zero for landscape mode and nonzero for portrait mode. If name is 
   empty or all blanks a default file name is selected; the user no longer has
   the option to plot to standard output. */
{
   FILE *headfile; /* File containing the plot header */
   char *namestr; /* Place to put a C string copy of name */
   int thechar;
#ifdef ENVIRON 
   char *headname; /* Name of PostScript header file returned by getenv() */
   char *getenv();
#endif ENVIRON
   /* First, try to open the appropriate file for output */
   if(blank(name,nlen)) { /* name is empty; query user for pathname */
     printf("Plot file name? --> ");
     namestr = (char *)calloc(1,128);
     scanf("%s",namestr);
   } else { /* name supplied */
     namestr = (char *)calloc(1,nlen+1);
     strncpy(namestr,name,nlen+1);
   }
   plotfile = fopen(namestr,"w");
   cfree(namestr);
   /* Check for successfully opened file */
   if((int)plotfile == NULL) {
     fprintf(stderr,"Can't open plot file!\n");
     exit(1);
   }
   /* Now, copy the PostScript header file to the output */
   /**** NOTE **** We should also look for an environment variable containing
       the default header file name.*/
#ifdef ENVIRON
   headname = getenv( ENVIRON );
   if((int)headname == NULL) 
     headfile = fopen("/private/swamy/geophysics_plotter/plothead.ps","r"); 
   else 
     headfile = fopen(headname,"r");
#else
   headfile = fopen("/private/swamy/geophysics_plotter/plothead.ps","r");
#endif ENVIRON
   if((int)headfile == NULL) {
     fprintf(stderr,"Can't open PostScript header file!\n");
     exit(1);
   }
   while((thechar = getc(headfile)) != EOF) putc(thechar,plotfile);
   fclose(headfile);
   /* Now, initialize the plot. */
   putc('\n',plotfile);
   fprintf(plotfile,"%d B\n",*idum);
   fact = 1.0;
   xx = yy = 0.0;
}

plot_(xrel,yrel,ipen)
float *xrel,*yrel; /* Addresses of FORTRAN REAL variables */
int *ipen;  /* Address of pen argument */
/* Moves pen to coordinates <*xrel,*yrel> in user unit relative to the 
   current origin using pen status <*ipen>.
  <*ipen> = 2  move with pen DOWN (draw a line)
           -2  move with pen DOWN and reorigin to <*xrel,*yrel>
            3  move with pen UP (no drawing)
           -3  move with pen UP and reorigin to <*xrel,*yrel>
      998,999  end plot file
 ***NOTE*** PLOTS must be called first !
*/
{
   int ixrel,iyrel;

   ixrel = (int)rint(*xrel * 1000.0);
   iyrel = (int)rint(*yrel * 1000.0);
   if(abs(*ipen) == 2) fprintf(plotfile,"%d %d l\n",ixrel,iyrel);
   if(abs(*ipen) == 3) fprintf(plotfile,"%d %d m\n",ixrel,iyrel);
   xx = *xrel;
   yy = *yrel;
   if(*ipen<0) {
     fprintf(plotfile,"r\n");
     xx = yy = 0.0;
   }
   if(*ipen>990) {
      fprintf(plotfile,"E\n");
      fclose(plotfile);
   }
}

poly_(xs,ys,n,stroke,fill)
float xs[],ys[];  /* Real arrays containing coordinates */
int *n; /* Number of points */
int *stroke; /* Boolean -- whether to stroke (outline) path */
float *fill; /* Fill level -- 0 (black) to 1 (white), or <0 for no fill */

/* This routine takes advantage of PostScript to draw an arbitrary object on
   the plot device. The XS and YS vectors contain *N points. A path is started
   at the first point given, and a line drawn to each point. The path is then
   filled (if FILL > 0.0) and stroked (if STROKE <> 0). */
{
   int ix,iy,i;
   
   fprintf(plotfile,"%%\ngsave\n%%\n");
   /* Save the graphics state, since we might be changing things */
   ix = (int)rint(xs[0] * 1000.0);
   iy = (int)rint(ys[0] * 1000.0);
   fprintf(plotfile,"%d %d m\n",ix,iy); /* Move to first point */
   for(i=1;i<*n;i++) {
     ix = (int)rint(xs[i] * 1000.0);
     iy = (int)rint(ys[i] * 1000.0);
     fprintf(plotfile,"%d %d n\n",ix,iy);
   }
   /* Now, the path has been created. We must now fill or stroke it. */
   fprintf(plotfile,"%%\n"); /* drivers: raw PostScript follows */
   if(*fill >=0.0) {
     fprintf(plotfile,"gsave\n"); /* save graphics state */
     fprintf(plotfile,"%f setgray fill ", (*fill>1.0) ? 1.0 : *fill );
     fprintf(plotfile,"grestore\n"); /* restore graphics state */
   }
   if(*stroke) 
     fprintf(plotfile,"stroke\n");
   fprintf(plotfile,"grestore\n%%\n"); /* restore state, done with raw PS */
}

shade_(xs,ys,gray,n,iw)
float xs[],ys[];  /* Real arrays containing coordinates */
float gray[];	  /* Array containing gray levels 0 (black) to 1 (white) */
int *n;           /* Number of points */
int *iw;          /* Width of line in 1/144 inches */

/* Routine to draw a curve XS, YS of length N.  The width of the curve (1/144
   inches) is IW and the gray shade at each point of the curve is 
   GRAY */
{
   int ix,iy,i,iwo,ig;
   
   fprintf(plotfile,"%%\ngsave 0 setlinecap\n%%\n");
   /* Save the graphics state, since we might be changing things */
   ix = (int)rint(xs[0] * 1000.0);
   iy = (int)rint(ys[0] * 1000.0);
   iwo = *iw + 11;
   fprintf(plotfile,"%d %d m\n",ix,iy); /* Move to first point */
   fprintf(plotfile,"%d p\n",iwo);      /* Set line width */
   for(i=1;i<*n;i++) {
     ix = (int)rint(xs[i] * 1000.0);
     iy = (int)rint(ys[i] * 1000.0);
     ig = (int)rint(99.*(gray[i-1]+gray[i])/2);
     fprintf(plotfile,"%d %d %d g\n",ix,iy,ig);
   }
   fprintf(plotfile,"%%\ngrestore\n%%\n"); /* restore state, done with raw PS */
}

raster_(idata,m,n,x0,y0,w,h)
short *idata; /* pointer to image data */
int *n,*m; /* rows&cols in image */
float *x0,*y0,*h,*w; /* upper left corner, height, width */

/* Plots a gray-scale image on the output device with 8 bits/pixel. */
{
   unsigned short *pixptr;
   int i,j,k = 0;
   unsigned char byte;
   
   fprintf(plotfile,"%d %d %d ",*m,*n,8);
   fprintf(plotfile,"%d %d %d %d i\n",(int)(*x0*1000),(int)((*y0-*h)*1000),(int)(*w*1000),(int)(*h*1000));
   for(i=0,pixptr = (unsigned short *)idata;i<((*n)*(*m));i++) {
     byte = (unsigned char)(*pixptr++);
     fprintf(plotfile,"%.2x",byte);
     k += 2;
     if (k>70) {
       fprintf(plotfile,"\n");
       k=0;
     }
   }
   fprintf(plotfile,"\n%%\n");
}

craster_(idata,m,n,x0,y0,w,h)
int *idata; /* points to image data - longwords of the form XXRRGGBB */
int *n,*m; /* rows&cols in image */
float *x0,*y0,*w,*h; /* upper left corner, width, height */

/* plots a color image on the output device with 24 bits per pixel */
{
   int i,j,k = 0;
   unsigned char *bytptr;
  
   fprintf(plotfile,"%d %d %d ",*m,*n,24);
   fprintf(plotfile,"%d %d %d %d c\n",(int)(*x0*1000),(int)((*y0-*h)*1000),(int)(*w*1000),(int)(*h*1000));
   for(i=0,bytptr = (unsigned char *)idata;i<((*m)*(*n));i++) {
     bytptr++; /* skip over first byte of longword */
     fprintf(plotfile,"%.2x",*bytptr++); /* red */
     fprintf(plotfile,"%.2x",*bytptr++); /* green */
     fprintf(plotfile,"%.2x",*bytptr++); /* blue */
     k += 6; 
     if(k>70) {
       fprintf(plotfile,"\n");
       k=0;
     }
   }
   fprintf(plotfile,"\n%%\n");
}

where_(xrel,yrel,factr)
float *xrel,*yrel; /* These point to the variables that hold returned coords */
float *factr; /* Returns the current scaling factor */
/* Returns in its parameters the current pen location relative to the current
   origin in user units, and the current scaling factor (1.0 --> true inches) */
{
   *xrel = (float)xx;
   *yrel = (float)yy;
   *factr = (float)fact;
}

newpen_(lpen)
int *lpen;  /* Designates pen to change to */
/* Sets a new pen color, line weight, or RGB color (for color terminals).
   LPEN is an 8-digit integer. The two lowest order digits determine pen
   number (for plotters) if less than 11, and for 11-99 they determine line
   weight. The six higher order digits determine RGB color, with two digits
   for each component. (RRGGBBPP) */
{
   fprintf(plotfile,"%8d p\n",*lpen);
}

advpag_(xpag)
float *xpag;
/* Behaves similarly to ADVPLT except that the pen is moved an extra XPAG
   units in the X direction -- most useful for things like fanfold pen plotters.
*/
{
   fprintf(plotfile,"%f N\n",*xpag);
}

advplt_()
/* Advances to the next plot frame. Usually this involves moving the pen to
   its maximum previous X value and a zero Y value (relative to the original 
   origin). */
{
   fprintf(plotfile,"P\n");
}

factor_(fctr)
float *fctr;  /* Address of new scaling factor */
/* Sets the new scaling factor for plotting (1.0 --> true inches). */
{
   fprintf(plotfile,"%f f\n",*fctr);
   fact = (double)(*fctr);
}

ginput_(xrel,yrel,key)
float *xrel,*yrel;
char *key;
/* This is used to find the location of the graphics crosshairs when a key is
   hit. Since all plotting is deferred, this merely returns the current pen
   position. */
{
   *xrel = (float)xx;
   *yrel = (float)yy;
   *key = ' ';
}


